(***********************************************************************)
(*                                                                     *)
(*                               TAToo                                 *)
(*                                                                     *)
(*                     Kim Nguyen, LRI UMR8623                         *)
(*                   Université Paris-Sud & CNRS                       *)
(*                                                                     *)
(*  Copyright 2010-2013 Université Paris-Sud and Centre National de la *)
(*  Recherche Scientifique. All rights reserved.  This file is         *)
(*  distributed under the terms of the GNU Lesser General Public       *)
(*  License, with the special exception on linking described in file   *)
(*  ../LICENSE.                                                        *)
(*                                                                     *)
(***********************************************************************)

INCLUDE "utils.ml"
open Format
open Misc

module Make (T : Tree.S) =
 struct

   module NodeSummary =
   struct
     (* Pack into an integer the result of the is_* and has_ predicates
        for a given node *)
     type t = int
     let dummy = -1
    (*
      4444444444443210
      4 -> kind
      3 -> is_left
      2 -> is_right
      1 -> has_left
      0 -> has_right
    *)

     let has_right (s : t) : bool =
       Obj.magic (s land 1)

     let has_left (s : t) : bool =
       Obj.magic ((s lsr 1) land 1)

     let is_right (s : t) : bool =
       Obj.magic ((s lsr 2) land 1)

     let is_left (s : t) : bool =
       Obj.magic ((s lsr 3) land 1)

     let kind (s : t) : Tree.NodeKind.t =
       Obj.magic (s lsr 4)

     let make is_left is_right has_left has_right kind =
       ((Obj.magic kind) lsl 4) lor
         ((int_of_bool is_left) lsl 3) lor
         ((int_of_bool is_right) lsl 2) lor
         ((int_of_bool has_left) lsl 1) lor
         (int_of_bool has_right)

   end

   type node_status = {
     sat : StateSet.t;
     unsat : StateSet.t;
     todo : Ata.TransList.t;
     summary : NodeSummary.t;
   }
(* Describe what is kept at each node for a run *)

   module NodeStatus = Hcons.Make(struct
     type t = node_status
     let equal c d =
       c == d ||
         c.sat == d.sat &&
         c.unsat == d.unsat &&
         c.todo == d.todo &&
         c.summary == d.summary

     let hash c =
       HASHINT4((c.sat.StateSet.id :> int),
                (c.unsat.StateSet.id :> int),
                (c.todo.Ata.TransList.id :> int),
                c.summary)
   end
   )

   let dummy_status =
     NodeStatus.make { sat = StateSet.empty;
                       unsat = StateSet.empty;
                       todo = Ata.TransList.nil;
                       summary = NodeSummary.dummy;
                     }


   type run = {
     tree : T.t ;
     (* The argument of the run *)
     auto : Ata.t;
     (* The automaton to be run *)
     status : NodeStatus.t array;
     (* A mapping from node preorders to NodeStatus *)
     unstable : Bitvector.t;
     (* A bitvector remembering whether a subtree is stable *)
     mutable redo : bool;
     (* A boolean indicating whether the run is incomplete *)
     mutable pass : int;
     (* The number of times this run was updated *)
     mutable cache2 : Ata.TransList.t Cache.N2.t;
     (* A cache from states * label to list of transitions *)
     mutable cache4 : NodeStatus.t Cache.N4.t;
   }

   let pass r = r.pass
   let stable r = not r.redo
   let auto r = r.auto
   let tree r = r.tree


   let dummy_trl =
     Ata.(TransList.cons
            (Transition.make
               (State.dummy,QNameSet.empty, Formula.false_))
            TransList.nil)

   let make auto tree =
     let len = T.size tree in
     {
       tree = tree;
       auto = auto;
       status = Array.create len dummy_status;
       unstable = Bitvector.create ~init:true len;
       redo = true;
       pass = 0;
       cache2 = Cache.N2.create dummy_trl;
       cache4 = Cache.N4.create dummy_status;
     }

   let get_status a i =
     if i < 0 then dummy_status else Array.get a i

   let unsafe_get_status a i =
     if i < 0 then dummy_status else Array.unsafe_get a i

IFDEF HTMLTRACE
  THEN
DEFINE TRACE(e) = (e)
  ELSE
DEFINE TRACE(e) = ()
END

   let html tree node i config msg =
     let config = config.NodeStatus.node in
     Html.trace (T.preorder tree node) i
       "node: %i<br/>%s<br/>sat: %a<br/>unsat: %a<br/>todo: %around: %i<br/>"
       (T.preorder tree node)
       msg
       StateSet.print config.sat
       StateSet.print config.unsat
       (Ata.TransList.print ~sep:"<br/>") config.todo i


   let debug msg tree node i config =
     let config = config.NodeStatus.node in
     eprintf
       "DEBUG:%s node: %i\nsat: %a\nunsat: %a\ntodo: %around: %i\n"
       msg
       (T.preorder tree node)
       StateSet.print config.sat
       StateSet.print config.unsat
       (Ata.TransList.print ~sep:"\n") config.todo i


   let get_trans cache2 auto tag states =
     let trs =
       Cache.N2.find cache2
         (tag.QName.id :> int) (states.StateSet.id :> int)
     in
     if trs == dummy_trl then
       let trs = Ata.get_trans auto tag states in
       (Cache.N2.add
          cache2
          (tag.QName.id :> int)
          (states.StateSet.id :> int) trs; trs)
     else trs



   let simplify_atom atom pos q { NodeStatus.node = status; _ } =
     if (pos && StateSet.mem q status.sat)
       || ((not pos) && StateSet.mem q status.unsat) then Ata.Formula.true_
     else if (pos && StateSet.mem q status.unsat)
         || ((not pos) && StateSet.mem q status.sat) then Ata.Formula.false_
     else atom


   let eval_form phi fcs nss ps ss summary =
     let open Ata in
         let rec loop phi =
           begin match Formula.expr phi with
             Boolean.True | Boolean.False -> phi
           | Boolean.Atom (a, b) ->
               begin
                 let open NodeSummary in
                     match a.Atom.node with
                     | Move (m, q) ->
                         let states = match m with
                           `First_child -> fcs
                         | `Next_sibling -> nss
                         | `Parent | `Previous_sibling -> ps
                         | `Stay -> ss
                         in simplify_atom phi b q states
                     | Is_first_child -> Formula.of_bool (b == is_left summary)
                     | Is_next_sibling -> Formula.of_bool (b == is_right summary)
                     | Is k -> Formula.of_bool (b == (k == kind summary))
                     | Has_first_child -> Formula.of_bool (b == has_left summary)
                     | Has_next_sibling -> Formula.of_bool (b == has_right summary)
               end
           | Boolean.And(phi1, phi2) -> Formula.and_ (loop phi1) (loop phi2)
           | Boolean.Or (phi1, phi2) -> Formula.or_  (loop phi1) (loop phi2)
           end
         in
         loop phi


   let eval_trans_aux cache4 fcs nss ps ss old_config =
     let { sat = old_sat;
           unsat = old_unsat;
           todo = old_todo;
           summary = old_summary } = old_config.NodeStatus.node
     in
     let sat, unsat, removed, kept, todo =
       Ata.TransList.fold
         (fun trs acc ->
           let q, lab, phi = Ata.Transition.node trs in
           let a_sat, a_unsat, a_rem, a_kept, a_todo = acc in
           if StateSet.mem q a_sat || StateSet.mem q a_unsat then acc else
             let new_phi =
               eval_form phi fcs nss ps old_config old_summary
             in
             if Ata.Formula.is_true new_phi then
               StateSet.add q a_sat, a_unsat, StateSet.add q a_rem, a_kept, a_todo
             else if Ata.Formula.is_false new_phi then
               a_sat, StateSet.add q a_unsat, StateSet.add q a_rem, a_kept, a_todo
             else
               let new_tr = Ata.Transition.make (q, lab, new_phi) in
               (a_sat, a_unsat, a_rem, StateSet.add q a_kept, (Ata.TransList.cons new_tr a_todo))
         ) old_todo (old_sat, old_unsat, StateSet.empty, StateSet.empty, Ata.TransList.nil)
     in
           (* States that have been removed from the todo list and not kept are now
              unsatisfiable *)
     let unsat = StateSet.union unsat (StateSet.diff removed kept) in
           (* States that were found once to be satisfiable remain so *)
     let unsat = StateSet.diff unsat sat in
     let new_config = NodeStatus.make { old_config.NodeStatus.node with sat; unsat; todo; } in
     new_config


   let eval_trans cache4 fcs nss ps ss =
     let fcsid = (fcs.NodeStatus.id :> int) in
     let nssid = (nss.NodeStatus.id :> int) in
     let psid = (ps.NodeStatus.id :> int) in
     let rec loop old_config =
       let oid = (old_config.NodeStatus.id :> int) in
       let res =
         let res = Cache.N4.find cache4 oid fcsid nssid psid in
         if res != dummy_status then res
         else
           let new_config = 
             eval_trans_aux cache4 fcs nss ps ss old_config
           in
           Cache.N4.add cache4 oid fcsid nssid psid new_config;
           new_config
       in
       if res == old_config then res else loop res
     in
     loop ss




  let top_down run =
    let tree = run.tree in
    let auto = run.auto in
    let status = run.status in
    let cache2 = run.cache2 in
    let cache4 = run.cache4 in
    let unstable = run.unstable in
    let rec loop node =
      let node_id = T.preorder tree node in
      if node == T.nil || not (Bitvector.get unstable node_id) then false else begin
        let parent = T.parent tree node in
        let fc = T.first_child tree node in
        let fc_id = T.preorder tree fc in
        let ns = T.next_sibling tree node in
        let ns_id = T.preorder tree ns in
        let tag = T.tag tree node in
        (* We enter the node from its parent *)

        let status0 =
          let c = unsafe_get_status status node_id in
          if c == dummy_status then
            (* first time we visit the node *)
            let ltrs = get_trans cache2 auto tag (Ata.get_states auto) in
            NodeStatus.make
              { sat = StateSet.empty;
                unsat = Ata.get_starting_states auto;
                todo = ltrs;
                summary = NodeSummary.make
                  (node == T.first_child tree parent) (* is_left *)
                  (node == T.next_sibling tree parent) (* is_right *)
                  (fc != T.nil) (* has_left *)
                  (ns != T.nil) (* has_right *)
                  (T.kind tree node) (* kind *)
              }
          else c
        in

        TRACE(html tree node _i config0 "Entering node");

        (* get the node_statuses for the first child, next sibling and parent *)
        let ps = unsafe_get_status status (T.preorder tree parent) in
        let fcs = unsafe_get_status status fc_id in
        let nss = unsafe_get_status status ns_id in
        (* evaluate the transitions with all this statuses *)
        let status1 = eval_trans cache4 fcs nss ps status0 in

        TRACE(html tree node _i config1 "Updating transitions");

        (* update the cache if the status of the node changed *)

        if status1 != status0 then status.(node_id) <- status1;
        (* recursively traverse the first child *)
        let unstable_left = loop fc in
        (* here we re-enter the node from its first child,
           get the new status of the first child *)
        let fcs1 = unsafe_get_status status fc_id in
        (* update the status *)
        let status2 = eval_trans cache4 fcs1 nss ps status1 in

        TRACE(html tree node _i config2 "Updating transitions (after first-child)");

        if status2 != status1 then status.(node_id) <- status2;
        let unstable_right = loop ns in
        let nss1 = unsafe_get_status status ns_id in
        let status3 = eval_trans cache4 fcs1 nss1 ps status2 in

        TRACE(html tree node _i config3 "Updating transitions (after next-sibling)");

        if status3 != status2 then status.(node_id) <- status3;

        let unstable_self =
          (* if either our left or right child is unstable or if we still have transitions
             pending, the current node is unstable *)
          unstable_left
          || unstable_right
          || Ata.TransList.nil != status3.NodeStatus.node.todo
        in
        Bitvector.unsafe_set unstable node_id unstable_self;
        TRACE((if not unstable_self then
            Html.finalize_node
              node_id
              _i
              Ata.(StateSet.intersect config3.Config.node.sat auto.selection_states)));
        unstable_self
      end
    in
    run.redo <- loop (T.root tree);
    run.pass <- run.pass + 1

(*
  let stats run =
    let count = ref 0 in
    let len = Bitvector.length run.unstable in
    for i = 0 to len - 1 do
      if not (Bitvector.unsafe_get run.unstable i) then
        incr count
    done;
    Logger.msg `STATS
      "%i nodes over %i were skipped in iteration %i (%.2f %%), redo is: %b"
      !count len run.pass (100. *. (float !count /. float len))
      run.redo


  let eval auto tree node =
    let len = T.size tree in
    let run = { config = Array.create len Ata.dummy_config;
                unstable = Bitvector.create ~init:true len;
                redo = true;
                pass = 0
              }
    in
    while run.redo do
      run.redo <- false;
      Ata.reset auto; (* prevents the .cache2 and .cache4 memoization tables from growing too much *)
      run.redo <- top_down_run auto tree node run;
      stats run;
      run.pass <- run.pass + 1;
    done;
    at_exit (fun () -> Logger.msg `STATS "%i iterations" run.pass);
    at_exit (fun () -> stats run);
    let r = get_results auto tree node run.config in

    TRACE(Html.gen_trace (module T : Tree.S with type t = T.t) (tree));

    r
*)

  let get_results run =
    let cache = run.status in
    let auto = run.auto in
    let tree = run.tree in
    let rec loop node acc =
      if node == T.nil then acc
      else
        let acc0 = loop (T.next_sibling tree node) acc in
        let acc1 = loop (T.first_child tree node) acc0 in

        if Ata.(
          StateSet.intersect
            cache.(T.preorder tree node).NodeStatus.node.sat
            (get_selecting_states auto)) then node::acc1
        else acc1
    in
    loop (T.root tree) []


  let get_full_results run =
    let cache = run.status in
    let auto = run.auto in
    let tree = run.tree in
    let res_mapper = Hashtbl.create MED_H_SIZE in
    let () =
      StateSet.iter
        (fun q -> Hashtbl.add res_mapper q [])
        (Ata.get_selecting_states auto)
    in
    let rec loop node =
      if node != T.nil then
        let () = loop (T.next_sibling tree node) in
        let () = loop (T.first_child tree node) in
        StateSet.iter
          (fun q ->
            try
              let acc = Hashtbl.find res_mapper q in
              Hashtbl.replace res_mapper q (node::acc)
            with
              Not_found -> ())
          cache.(T.preorder tree node).NodeStatus.node.sat
    in
    loop (T.root tree);
    StateSet.fold
      (fun q acc -> (q, Hashtbl.find res_mapper q)::acc)
      (Ata.get_selecting_states auto) []

  let prepare_run run list =
    let tree = run.tree in
    let auto = run.auto in
    let status = run.status in
    let cache2 = run.cache2 in
    List.iter (fun node ->
      let parent = T.parent tree node in
      let fc = T.first_child tree node in
      let ns = T.next_sibling tree node in
      let tag = T.tag tree node in

      let status0 =
        NodeStatus.make
          { sat = Ata.get_starting_states auto;
            unsat = StateSet.empty;
            todo = get_trans cache2 auto tag (Ata.get_states auto);
            summary = NodeSummary.make
              (node == T.first_child tree parent) (* is_left *)
              (node == T.next_sibling tree parent) (* is_right *)
              (fc != T.nil) (* has_left *)
              (ns != T.nil) (* has_right *)
              (T.kind tree node) (* kind *)
          }
      in
      let node_id = T.preorder tree node in
      status.(node_id) <- status0) list


  let eval full auto tree nodes =
    let run = make auto tree in
    prepare_run run nodes;
    while run.redo do
      top_down run
    done;
    if full then `Full (get_full_results run)
    else `Normal (get_results run)


  let full_eval auto tree nodes =
    match eval true auto tree nodes with
      `Full l -> l
    | _ -> assert false

  let eval auto tree nodes =
    match eval false auto tree nodes with
      `Normal l -> l
    | _ -> assert false

end
